import wx
from .amulet_ui import AmuletUI
import sys
import locale


class AmuletApp(wx.App):
    def OnInit(self):
        self._frame = AmuletUI(None)
        self.SetTopWindow(self._frame)
        self._frame.Show()
        return True

    def InitLocale(self):
        # https://discuss.wxpython.org/t/what-is-wxpython-doing-to-the-locale-to-makes-pandas-crash/34606/18
        if sys.version_info[:2] >= (3, 8):
            super().InitLocale()
        else:
            self.ResetLocale()
            lang, enc = locale.getlocale()
            if lang is None:
                self._initial_locale = wx.Locale(wx.LANGUAGE_DEFAULT)



from typing import Tuple, Optional, Any, TYPE_CHECKING
import wx
import weakref
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.history.history_manager import ObjectHistoryManager
from amulet.api.history import Changeable

from amulet_map_editor import log

if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas

BoxType = Tuple[Tuple[int, int, int], Tuple[int, int, int]]  # min and max positions


_SelectionChangeEventType = wx.NewEventType()
EVT_SELECTION_CHANGE = wx.PyEventBinder(_SelectionChangeEventType)


class SelectionChangeEvent(wx.PyEvent):
    """Run when the selection is changed by third party code."""

    def __init__(self):
        wx.PyEvent.__init__(self, eventType=_SelectionChangeEventType)


class SelectionManager(Changeable):
    """A class containing the raw representation of the selection with methods to access and change it."""

    def __init__(self, canvas: "EditCanvas"):
        super().__init__()
        self._selection_corners: Tuple[BoxType, ...] = ()
        self._selection_group: SelectionGroup = SelectionGroup()
        self._canvas = weakref.ref(canvas)

        self._timer = wx.Timer(canvas)

    @property
    def canvas(self) -> "EditCanvas":
        return self._canvas()

    def bind_events(self):
        """Set up all events required to run."""
        self.canvas.Bind(wx.EVT_TIMER, self._create_undo_point, self._timer)
        self.canvas.Bind(wx.EVT_WINDOW_DESTROY, self._on_destroy, self.canvas)

    def _create_undo_point(self, evt):
        self.canvas.create_undo_point(False, True)
        evt.Skip()

    def _start_undo_point(self):
        """Start a timer to create an undo point after a period of time.
        If this is called again before the timer runs then the last call will not happen."""
        self._timer.StartOnce(400)

    def _on_destroy(self, evt):
        self._timer.Stop()
        evt.Skip()

    @property
    def selection_corners(
        self,
    ) -> Tuple[Tuple[Tuple[int, int, int], Tuple[int, int, int]], ...]:
        """Get the minimum and maximum points of each selection
        :return: The minimum and maximum points of each selection
        """
        return self._selection_corners

    @selection_corners.setter
    def selection_corners(
        self,
        selection_corners: Tuple[
            Tuple[Tuple[int, int, int], Tuple[int, int, int]], ...
        ],
    ):
        """Set the minimum and maximum points of each selection
        Will create events that allow the program to update.

        :param selection_corners: The minimum and maximum points of each selection
        :return:
        """
        self.set_selection_corners(selection_corners)
        self._start_undo_point()

    def set_selection_corners(
        self,
        selection_corners: Tuple[
            Tuple[Tuple[int, int, int], Tuple[int, int, int]], ...
        ],
    ):
        """Set the minimum and maximum points of each selection
        Note this method will not trigger the history logic.
        You may instead want the selection_corners setter method.

        :param selection_corners: The minimum and maximum points of each selection
        :return:
        """
        selections = []
        for points in selection_corners:
            if len(points) == 2 and all(len(point) == 3 for point in points):
                selections.append(
                    tuple(tuple(int(p) for p in point) for point in points)
                )
            else:
                log.error(
                    f"selection_corners must be of the format Tuple[Tuple[Tuple[int, int, int], Tuple[int, int, int]], ...]"
                )

        self.changed = True
        self._selection_corners = tuple(selections)
        self._selection_group = SelectionGroup(
            [SelectionBox(*box) for box in self._selection_corners]
        )
        wx.PostEvent(self._canvas(), SelectionChangeEvent())

    @property
    def selection_group(self) -> SelectionGroup:
        """Get the selection as a `SelectionGroup`
        :return: `SelectionGroup`
        """
        return self._selection_group

    @selection_group.setter
    def selection_group(self, selection_group: SelectionGroup):
        """Set the selection from a `SelectionGroup` class
        Will create events that allow the program to update.

        :param selection_group: The `SelectionGroup` to pull the data from
        :return:
        """
        self.set_selection_group(selection_group)
        self._start_undo_point()

    def set_selection_group(self, selection_group: SelectionGroup):
        """Set the selection from a `SelectionGroup` class
        Note this method will not trigger the history logic.
        You may instead want the selection_group setter method.

        :param selection_group: The `SelectionGroup` to pull the data from
        :return:
        """
        self.changed = True
        self._selection_corners = [
            (box.min, box.max) for box in selection_group.selection_boxes
        ]
        self._selection_group = selection_group
        wx.PostEvent(self._canvas(), SelectionChangeEvent())


class SelectionHistoryManager(ObjectHistoryManager):
    def __init__(self, selection_manager: SelectionManager):
        super().__init__(selection_manager)

    @property
    def value(self) -> SelectionManager:
        return self._value

    def _unpack_value(self, value: Optional[Any]):
        self.value.set_selection_corners(value)

    def _pack_value(self, value: SelectionManager) -> Optional[Any]:
        return value.selection_corners



from typing import Tuple
import wx

import weakref
import numpy
import math

from amulet_map_editor.api.opengl.camera import Camera
from amulet_map_editor.api.opengl.matrix import rotation_matrix_xy
from amulet_map_editor.programs.edit.api.key_config import (
    KeybindGroup,
    ACT_MOVE_UP,
    ACT_MOVE_DOWN,
    ACT_MOVE_FORWARDS,
    ACT_MOVE_BACKWARDS,
    ACT_MOVE_LEFT,
    ACT_MOVE_RIGHT,
    ACT_BOX_CLICK,
)
from amulet_map_editor.api.wx.util.button_input import (
    ButtonInput,
    InputPressEvent,
    EVT_INPUT_PRESS,
    InputReleaseEvent,
    EVT_INPUT_RELEASE,
    InputHeldEvent,
    EVT_INPUT_HELD,
)


_MoveActions = {
    ACT_MOVE_UP,
    ACT_MOVE_DOWN,
    ACT_MOVE_FORWARDS,
    ACT_MOVE_BACKWARDS,
    ACT_MOVE_LEFT,
    ACT_MOVE_RIGHT,
}


class NudgeButton(wx.Button):
    """A button that catches actions when pressed."""

    def __init__(
        self,
        parent: wx.Window,
        camera: Camera,
        keybinds: KeybindGroup,
        label: str,
        tooltip: str,
    ):
        super().__init__(parent, label=label)
        self.SetToolTip(tooltip)
        self._camera = weakref.ref(camera)
        self._buttons = ButtonInput(self)
        self._buttons.register_actions(keybinds)
        self._buttons.bind_events()  # this is fine here because we are binding to a custom button not the canvas.
        self.Bind(EVT_INPUT_PRESS, self._on_down)
        self.Bind(EVT_INPUT_RELEASE, self._on_up)
        self.Bind(EVT_INPUT_HELD, self._on_held)
        self._listen = False
        self._timeout = 10

    @property
    def camera(self) -> Camera:
        return self._camera()

    def enable(self):
        self._buttons.enable()

    def disable(self):
        self._buttons.disable()

    def _on_down(self, evt: InputPressEvent):
        if evt.action_id == ACT_BOX_CLICK:
            self._listen = True
        elif evt.action_id in _MoveActions:
            self._timeout = 10

    def _on_up(self, evt: InputReleaseEvent):
        if evt.action_id == ACT_BOX_CLICK:
            self._listen = False

    def _on_held(self, evt: InputHeldEvent):
        if self._listen:
            if self._timeout == 0 or self._timeout == 10:
                x = y = z = 0
                if ACT_MOVE_LEFT in evt.action_ids:
                    x += 1
                if ACT_MOVE_RIGHT in evt.action_ids:
                    x -= 1
                if ACT_MOVE_UP in evt.action_ids:
                    y += 1
                if ACT_MOVE_DOWN in evt.action_ids:
                    y -= 1
                if ACT_MOVE_FORWARDS in evt.action_ids:
                    z += 1
                if ACT_MOVE_BACKWARDS in evt.action_ids:
                    z -= 1
                if any((x, y, z)):
                    self._move(self._rotate((x, y, z)))
            if self._timeout:
                self._timeout -= 1

    def _rotate(self, offset: Tuple[int, int, int]) -> Tuple[int, int, int]:
        x, y, z = offset
        ry = self.camera.rotation[0]
        x, y, z, _ = (
            numpy.round(
                numpy.matmul(
                    rotation_matrix_xy(0, -math.radians(round(ry / 90) * 90)),
                    (x, y, z, 0),
                )
            )
            .astype(int)
            .tolist()
        )
        return x, y, z

    def _move(self, offset: Tuple[int, int, int]):
        pass



import wx
import math

from amulet.api.data_types import BlockCoordinates, FloatTriplet

# The location changed
_ObjectLocationChangeEventType = wx.NewEventType()
EVT_LOCATION_CHANGE = wx.PyEventBinder(_ObjectLocationChangeEventType)


class LocationChangeEvent(wx.PyEvent):
    """Run when the camera has moved or rotated."""

    def __init__(self, location: BlockCoordinates):
        wx.PyEvent.__init__(self, eventType=_ObjectLocationChangeEventType)
        self._location = location

    @property
    def location(self) -> BlockCoordinates:
        """The location of the object. (x, y, z)"""
        return self._location


# The rotation changed
_ObjectRotationChangeEventType = wx.NewEventType()
EVT_ROTATION_CHANGE = wx.PyEventBinder(_ObjectRotationChangeEventType)


class RotationChangeEvent(wx.PyEvent):
    """Run when the camera has moved or rotated."""

    def __init__(self, rotation: FloatTriplet):
        wx.PyEvent.__init__(self, eventType=_ObjectRotationChangeEventType)
        self._rotation = rotation

    @property
    def rotation(self) -> FloatTriplet:
        """The rotation of the object. (x, y, z)."""
        return self._rotation

    @property
    def rotation_radians(self) -> FloatTriplet:
        """The rotation of the object. (x, y, z)."""
        return tuple(math.radians(r) for r in self._rotation)


# The scale changed
_ObjectScaleChangeEventType = wx.NewEventType()
EVT_SCALE_CHANGE = wx.PyEventBinder(_ObjectScaleChangeEventType)


class ScaleChangeEvent(wx.PyEvent):
    """Run when the camera has moved or rotated."""

    def __init__(self, scale: FloatTriplet):
        wx.PyEvent.__init__(self, eventType=_ObjectScaleChangeEventType)
        self._scale = scale

    @property
    def scale(self) -> FloatTriplet:
        """The scale of the object. (x, y, z)."""
        return self._scale


# The location, rotation or scale changed
_ObjectTransformEventType = wx.NewEventType()
EVT_TRANSFORM_CHANGE = wx.PyEventBinder(_ObjectTransformEventType)


class TransformChangeEvent(wx.PyEvent):
    """Run when the camera has moved or rotated."""

    def __init__(
        self, location: BlockCoordinates, rotation: FloatTriplet, scale: FloatTriplet
    ):
        wx.PyEvent.__init__(self, eventType=_ObjectTransformEventType)
        self._location = location
        self._rotation = rotation
        self._scale = scale

    @property
    def location(self) -> BlockCoordinates:
        """The location of the object. (x, y, z)"""
        return self._location

    @property
    def rotation(self) -> FloatTriplet:
        """The rotation of the object. (x, y, z)."""
        return self._rotation

    @property
    def rotation_radians(self) -> FloatTriplet:
        """The rotation of the object. (x, y, z)."""
        return tuple(math.radians(r) for r in self._rotation)

    @property
    def scale(self) -> FloatTriplet:
        """The scale of the object. (x, y, z)."""
        return self._scale



# Fixed function example plugin
# The fixed function operation pipeline works in much the same way as MCEdit Unified filters with some modifications
# You define a function and it will appear in the UI for you to run


from amulet.api.selection import SelectionGroup
from amulet.api.level import BaseLevel
from amulet.api.data_types import Dimension


# for those that are new to python 3 the thing after the colon is the object type that the variable should be
def operation(
    world: BaseLevel, dimension: Dimension, selection: SelectionGroup, options: dict
):
    # world is the object that contains all the data related to the world
    # dimension in a string used to identify the currently loaded dimension. It can be used to access the right dimension from the world
    # selection is an object describing the selections made by the user. It is possible there may not be any boxes selected
    # options will be explored in further examples
    pass


export = {  # This is what the program will actually look for. It describes how the operation will work
    "name": "Fixed Function Pipeline Example 1",  # the name of the plugin
    "operation": operation,  # the actual function to call when running the plugin
}



import wx
from typing import TYPE_CHECKING
import traceback

import amulet
from amulet.api.errors import LoaderNoneMatched

from amulet_map_editor import log
from amulet_map_editor.api.wx.ui.traceback_dialog import TracebackDialog
from amulet_map_editor.programs.edit.api.ui.tool import DefaultBaseToolUI
from amulet_map_editor.programs.edit.api.behaviour import StaticSelectionBehaviour


if TYPE_CHECKING:
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class ImportTool(wx.BoxSizer, DefaultBaseToolUI):
    def __init__(self, canvas: "EditCanvas"):
        wx.BoxSizer.__init__(self, wx.VERTICAL)
        DefaultBaseToolUI.__init__(self, canvas)

        self._selection = StaticSelectionBehaviour(self.canvas)

        self._open_file_button = wx.Button(canvas, label="Import File")
        self._open_file_button.Bind(wx.EVT_BUTTON, self._on_open_file)
        self.AddStretchSpacer()
        self.Add(self._open_file_button, flag=wx.ALL, border=10)
        self.AddStretchSpacer()

    @property
    def name(self) -> str:
        return "Import"

    def bind_events(self):
        super().bind_events()
        self._selection.bind_events()

    def enable(self):
        super().enable()
        self._selection.update_selection()

    def disable(self):
        super().disable()

    def _on_open_file(self, evt):
        with wx.FileDialog(
            self.canvas,
            "Open a Minecraft data file",
            wildcard="|".join(
                [  # TODO: Automatically load these from the FormatWrapper classes.
                    "All files (*.construction;*.mcstructure;*.schematic)|*.construction;*.mcstructure;*.schematic",
                    "Construction file (*.construction)|*.construction",
                    "Bedrock mcstructure file (*.mcstructure)|*.mcstructure",
                    "Legacy Schematic file (*.schematic)|*.schematic",
                    "Sponge Schematic file (*.schem)|*.schem",
                ]
            ),
            style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST,
        ) as fileDialog:
            if fileDialog.ShowModal() == wx.ID_CANCEL:
                return
            else:
                pathname = fileDialog.GetPath()
        try:
            level = amulet.load_level(pathname)
        except LoaderNoneMatched:
            msg = f"Could not find a matching loader for {pathname}."
            log.error(msg)
            wx.MessageBox(msg)
        except Exception as e:
            log.error(f"Could not open {pathname}.", exc_info=True)
            dialog = TracebackDialog(
                self.canvas,
                f"Could not open {pathname}.",
                str(e),
                traceback.format_exc(),
            )
            dialog.ShowModal()
            dialog.Destroy()
        else:
            self.canvas.paste(level, level.dimensions[0])



from OpenGL.GL import (
    glBindTexture,
    GL_TEXTURE_2D,
    glTexImage2D,
    GL_RGBA,
    GL_UNSIGNED_BYTE,
    glGenTextures,
    glTexParameteri,
    GL_TEXTURE_MIN_FILTER,
    GL_NEAREST,
    GL_CLAMP_TO_EDGE,
    GL_TEXTURE_MAG_FILTER,
    GL_TEXTURE_WRAP_S,
    GL_TEXTURE_WRAP_T,
)
from typing import Generator, Any, Tuple, Dict, Optional
import struct
import hashlib
import os
import json
from PIL import Image
import numpy
import glob

from minecraft_model_reader.api.resource_pack.base import BaseResourcePackManager
from minecraft_model_reader import BlockMesh
import PyMCTranslate
from amulet.api.block import Block

from amulet_map_editor import log
from amulet_map_editor.api.opengl import textureatlas


class OpenGLResourcePack:
    """This class will take a minecraft_model_reader resource pack and
    load all of the textures into a texture atlas."""

    _translator: PyMCTranslate.Version
    _block_models: Dict[Block, BlockMesh]
    _texture_bounds: Dict[Any, Tuple[float, float, float, float]]
    _image: Optional[numpy.ndarray]
    _image_width: int
    _image_height: int
    _gl_textures: Dict[str, int]

    def __init__(
        self, resource_pack: BaseResourcePackManager, translator: PyMCTranslate.Version
    ):
        self._resource_pack = resource_pack
        self._translator = translator
        self._block_models: Dict[Block, BlockMesh] = {}

        self._texture_bounds: Dict[str, Tuple[float, float, float, float]] = {}
        self._image: Optional[Image.Image] = None
        self._image_width: int = 0
        self._image_height: int = 0

        self._gl_textures: Dict[str, int] = {}

    def get_atlas_id(self, context_id: str) -> int:
        """Get the opengl texture id of the atlas for a given context."""
        if context_id not in self._gl_textures:
            if self._image is None:
                raise Exception(
                    "OpenGLResourcePack.setup() needs to be run before accessing a texture."
                )
            self._setup_texture(context_id)
        return self._gl_textures[context_id]

    def get_texture_path(self, namespace: Optional[str], relative_path: str):
        """Get the absolute path of the image from the relative components.
        Useful for getting the id of textures for hard coded textures not connected to a resource pack."""
        return self._resource_pack.get_texture_path(namespace, relative_path)

    def texture_bounds(self, texture_path: str) -> Tuple[float, float, float, float]:
        """Get the bounding box of a given texture path."""
        if texture_path in self._texture_bounds:
            return self._texture_bounds[texture_path]
        else:
            return self._texture_bounds[self._resource_pack.missing_no]

    @property
    def translator(self) -> PyMCTranslate.Version:
        """The translator used to convert the universal blocks into the required version for the resource pack."""
        return self._translator

    def setup(self) -> Generator[float, None, None]:
        """Create and bind the atlas texture."""
        if self._image is None:
            cache_id = struct.unpack(
                "H",
                hashlib.sha1(
                    "".join(self._resource_pack.pack_paths).encode("utf-8")
                ).digest()[:2],
            )[0]

            atlas: Image.Image

            mod_time = max(
                os.stat(path).st_mtime
                for pack in self._resource_pack.pack_paths
                for path in glob.glob(os.path.join(pack, "**", "*.*"), recursive=True)
            )

            cache_dir = os.path.join(".", "cache", "resource_pack")
            img_path = os.path.join(cache_dir, f"{cache_id}.png")
            bounds_path = os.path.join(cache_dir, f"{cache_id}.json")
            try:
                with open(bounds_path) as f:
                    cache_mod_time, bounds = json.load(f)
                if mod_time != cache_mod_time:
                    raise Exception(
                        "The resource packs have changed since last merging."
                    )
                atlas = Image.open(img_path)
            except:
                atlas_iter = textureatlas.create_atlas_iter(
                    self._resource_pack.textures
                )
                try:
                    while True:
                        yield next(atlas_iter)
                except StopIteration as e:
                    (
                        atlas,
                        bounds,
                    ) = e.value
                    os.makedirs(cache_dir, exist_ok=True)
                    atlas.save(img_path)
                    with open(bounds_path, "w") as f:
                        json.dump((mod_time, bounds), f)

            self._image_width, self._image_height = atlas.size
            self._image = numpy.array(atlas).astype(numpy.uint8).ravel()
            self._texture_bounds = bounds

    def _setup_texture(self, context_id: str):
        """Set up the texture for a given context"""
        gl_texture = self._gl_textures[context_id] = glGenTextures(
            1
        )  # Create the texture location
        glBindTexture(GL_TEXTURE_2D, gl_texture)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)

        glBindTexture(GL_TEXTURE_2D, gl_texture)
        glTexImage2D(
            GL_TEXTURE_2D,
            0,
            GL_RGBA,
            self._image_width,
            self._image_height,
            0,
            GL_RGBA,
            GL_UNSIGNED_BYTE,
            self._image,
        )
        glBindTexture(GL_TEXTURE_2D, 0)
        log.info("Finished setting up texture atlas in OpenGL")

    def get_block_model(self, universal_block: Block) -> BlockMesh:
        """Get the BlockMesh class for a given universal Block.
        The Block will be translated to the version format using the
        previously specified translator."""
        if universal_block not in self._block_models:
            version_block = self._translator.block.from_universal(
                universal_block.base_block
            )[0]
            if universal_block.extra_blocks:
                for block_ in universal_block.extra_blocks:
                    version_block += self._translator.block.from_universal(block_)[0]

            self._block_models[universal_block] = self._resource_pack.get_block_model(
                version_block
            )

        return self._block_models[universal_block]



import os
import sys
from typing import Optional
from types import ModuleType
import importlib.util

from amulet_map_editor.programs.edit.plugins.operations import stock_plugins


STOCK_PLUGINS_DIR: str = stock_plugins.__path__[0]
STOCK_PLUGINS_NAME: str = stock_plugins.__name__
CUSTOM_PLUGINS_DIR = os.path.abspath("plugins")


def load_module(path: str) -> Optional[ModuleType]:
    """Load a module or package from the file path.
    If it is a module (a file ending .py) then path must be the path to this file.
    If it is a package (a directory optionally containing an __init__.py file) then the path of the directory must be given.

    :param path: The file path of the module/package.
    :return: module if loaded else None
    """
    if os.path.isfile(path):
        spec = importlib.util.spec_from_file_location(path, path)
        if spec is None:
            raise ImportError(f"Could not import {path}")
        mod = importlib.util.module_from_spec(spec)
    elif os.path.isdir(path):
        spec = importlib.util.spec_from_file_location(
            path,
            os.path.join(path, "__init__.py"),
        )
        if spec is None:
            raise ImportError(f"Could not import {path}")
        mod = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = mod
    else:
        return

    spec.loader.exec_module(mod)
    return mod



from __future__ import annotations

from typing import Tuple
import urllib.request
import threading
import json
import re
import webbrowser

import wx

from amulet_map_editor import lang

URL = "http://api.github.com/repos/Amulet-Team/Amulet-Map-Editor/releases"
VERSION_REGEX = re.compile(
    r"^v?(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<patch>\d+))?(\.(?P<bugfix>\d+))?(b(?P<beta>\d+))?$"
)

_EVT_UPDATE_CHECK = wx.NewEventType()
EVT_UPDATE_CHECK = wx.PyEventBinder(_EVT_UPDATE_CHECK, 1)


def get_version(version_string: str) -> Tuple[bool, Tuple[int, ...]]:
    """Parse the version into a more usable format

    :param version_string: The version string. Eg 1.2 or 1.2.3.4 or 1.2.3.4b0
    :return: (bool(full_version), (major, minor, patch, bugfix, <beta>)) beta will not exist if it is not a beta
    """
    version_match = VERSION_REGEX.match(version_string)
    if version_match:
        v = version_match.groupdict()
        version = (
            int(v["major"]),
            int(v["minor"]),
            int(
                v["patch"] or 0,
            ),
            int(
                v["bugfix"] or 0,
            ),
        )
        if v["beta"] is None:
            # full release
            return (True, version)
        else:
            # beta release
            return (False, version + (int(v["beta"]),))

    raise Exception(f"Invalid version string {version_string}")


class UpdateEvent(wx.PyCommandEvent):
    def __init__(self, etype, eid, new_version: str):
        wx.PyCommandEvent.__init__(self, etype, eid)
        self._new_version = new_version

    def GetVersion(self) -> str:
        return self._new_version


class CheckForUpdate(threading.Thread):
    def __init__(self, url, current_version, parent):
        threading.Thread.__init__(self)
        self.url = url
        self.current_version = current_version
        self._parent = parent
        self._new_version = None

    def run(self):
        try:
            conn = urllib.request.urlopen(self.url, timeout=5)
            data = conn.read()
            data = json.loads(data)
            try:
                current_version = get_version(self.current_version)
            except Exception:
                return

            for release in data:
                try:
                    release_version = get_version(release["tag_name"])
                except Exception:
                    continue

                if current_version[0] == release_version[0]:
                    # if they are both full releases or betas
                    if release_version[1] > current_version[1]:
                        self._new_version = release["tag_name"]
                        break
                elif not current_version[0]:
                    # if current is beta
                    if release_version[0]:
                        release_version = (
                            release_version[0],
                            release_version[1] + (float("inf"),),
                        )
                    if release_version[1] > current_version[1]:
                        self._new_version = release["tag_name"]
                        break

            if self._new_version:
                evt = UpdateEvent(_EVT_UPDATE_CHECK, -1, self._new_version)
                wx.PostEvent(self._parent, evt)
        except Exception:
            pass


class UpdateDialog(wx.Dialog):
    def __init__(self, parent, current_version: str, new_version: str):
        wx.Dialog.__init__(self, parent)

        sizer_1 = wx.BoxSizer(wx.VERTICAL)

        static_text_1 = wx.StaticText(
            self, label=lang.get("update_check.newer_version_released")
        )
        static_text_2 = wx.StaticText(
            self, label=f"{lang.get('update_check.new_version')} v{new_version}"
        )
        static_text_3 = wx.StaticText(
            self, label=f"{lang.get('update_check.current_version')} v{current_version}"
        )

        sizer_1.Add(static_text_1, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        sizer_1.Add(static_text_2, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)
        sizer_1.Add(static_text_3, 0, wx.ALL | wx.ALIGN_CENTER_HORIZONTAL, 5)

        sizer_2 = wx.BoxSizer(wx.HORIZONTAL)

        update_button = wx.Button(self, label=lang.get("update_check.update"))
        ok_button = wx.Button(self, label=lang.get("update_check.ok"))

        sizer_2.Add(update_button, 0, wx.ALL, 5)
        sizer_2.Add((0, 0), 1, wx.EXPAND, 5)
        sizer_2.Add(ok_button, 0, wx.ALL, 5)

        sizer_1.Add(sizer_2, 1, wx.EXPAND, 5)

        self.SetSizerAndFit(sizer_1)

        self.Centre()

        update_button.Bind(
            wx.EVT_BUTTON, lambda evt: self.goto_download_page(new_version, evt)
        )
        ok_button.Bind(wx.EVT_BUTTON, lambda evt: self.Close())

    @staticmethod
    def goto_download_page(new_version, _):
        webbrowser.open(
            f"https://github.com/Amulet-Team/Amulet-Map-Editor/releases/tag/{new_version}"
        )


def show_update_window(parent, current_version: str, evt: UpdateEvent):
    UpdateDialog(parent, current_version, evt.GetVersion()).ShowModal()


def check_for_update(version, listening_parent):
    update_thread = CheckForUpdate(URL, version, listening_parent)
    update_thread.start()



import weakref
import wx


class WindowContainer:
    """A helper class to store a reference to a wx.Window.
    If a window is hard referenced there may be cyclic references leading to memory leaks.
    Subclass this class if you intend to store a reference to a window."""

    def __init__(self, window: wx.Window):
        self._window = weakref.ref(window)

    @property
    def window(self) -> wx.Window:
        return self._window()



from amulet_map_editor import log


class BaseOperationLoader:
    """A class to load and wrap an operation"""

    def __init__(self, identifier: str, export_dict: dict):
        """
        Set up the OperationLoader.
        :param identifier: The path to the operation file. This may be a file or directory depending on if it is a module or package.
        """
        self._identifier = identifier
        self._name = ""
        self._is_valid = False
        self._load(export_dict)

    def _load(self, export_dict: dict):
        """load the operation."""
        if not isinstance(export_dict, dict):
            log.error(f"The export for {self.identifier} is not a dictionary.")
            return
        if "name" in export_dict and isinstance(export_dict["name"], str):
            self._name = export_dict["name"]
            self._setup(export_dict)
        else:
            log.error(f"Missing or invalid name in {self.identifier}")

    def _setup(self, export_dict: dict):
        """Parse the export dictionary and setup as required."""
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        """Run the actual operation."""
        raise NotImplementedError

    @property
    def is_valid(self) -> bool:
        """Is the module/package a valid operation."""
        return self._is_valid

    @property
    def identifier(self) -> str:
        """The identifier for this operation.
        This is formed of the path and optionally a number if there are multiple operations in the file."""
        return self._identifier

    @property
    def name(self) -> str:
        """The name of the operation."""
        return self._name



import numpy
from typing import TYPE_CHECKING, Tuple
import wx

from amulet_map_editor.api.wx.ui.base_select import EVT_PICK
from amulet_map_editor.api.wx.ui.simple import SimpleDialog
from amulet_map_editor.api.wx.ui.block_select import BlockDefine
from amulet_map_editor.programs.edit.api.operations import DefaultOperationUI
from amulet_map_editor.api import image

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas

MODES = {
    "Overlay": "Overlay the block on the existing. If the first block is air it replaces the first block otherwise it sets the second block.",
    "Underlay": "Underlay the block on the existing. If the first block is not air it moves the first block to the second block. It then replaces the first block.",
    "Normal Fill": "Replace all selected blocks with the chosen block.",
    "Game Fill": "Fill like in game. Moves water blocks to the second block and replaces the first block. Useful for setting blocks in existing water.",
    "Set First": "Set the first block leaving the second block as it was. If you want to change the first block without modifying the second block.",
    "Set Second": 'Set the second block leaving the first block as it was. This can cause issues if the first block is air. You might prefer "Overlay"',
}


class Waterlog(wx.Panel, DefaultOperationUI):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        wx.Panel.__init__(self, parent)
        DefaultOperationUI.__init__(self, parent, canvas, world, options_path)
        self.Freeze()

        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)

        options = self._load_options({})

        top_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self._sizer.Add(top_sizer, 0, wx.EXPAND | wx.ALL, 5)

        help_button = wx.BitmapButton(
            self, bitmap=image.icon.tablericons.help.bitmap(22, 22)
        )
        top_sizer.Add(help_button)

        def on_button(evt):
            dialog = SimpleDialog(self, "Extra block help.")
            text = wx.TextCtrl(
                dialog,
                value="Blocks in the newer versions of Minecraft support having two blocks in the same location.\n"
                "This is how the game is able to have water and blocks like fences at the same location.\n"
                "In the example of waterlogged fences the fence is the first block and the water is the second. Unless it is water the second block is usually just visual.\n"
                "In Java currently the second block is strictly water but in Bedrock there is no limit on what the second block can be.\n"
                "It is not possible to set non-water second blocks in the game but this operation enables the use of that feature.\n"
                "There are a number of different modes which can be selected at the top. A description of how it works will appear.",
                style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_BESTWRAP,
            )
            dialog.sizer.Add(text, 1, wx.EXPAND)
            dialog.ShowModal()
            evt.Skip()

        help_button.Bind(wx.EVT_BUTTON, on_button)

        self._mode = wx.Choice(self, choices=list(MODES.keys()))
        self._mode.SetSelection(0)
        top_sizer.Add(self._mode, 1, wx.EXPAND | wx.LEFT, 5)
        self._mode.Bind(wx.EVT_CHOICE, self._on_mode_change)

        self._mode_description = wx.TextCtrl(
            self, style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_BESTWRAP
        )
        self._sizer.Add(self._mode_description, 0, wx.EXPAND | wx.LEFT | wx.RIGHT, 5)

        self._mode_description.SetLabel(
            MODES[self._mode.GetString(self._mode.GetSelection())]
        )
        self._mode_description.Fit()

        self._block_define = BlockDefine(
            self,
            world.level_wrapper.translation_manager,
            wx.VERTICAL,
            *(options.get("fill_block_options", []) or [world.level_wrapper.platform]),
            show_pick_block=True
        )
        self._block_define.Bind(EVT_PICK, self._on_pick_block_button)
        self._sizer.Add(self._block_define, 1, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self._run_button = wx.Button(self, label="Run Operation")
        self._run_button.Bind(wx.EVT_BUTTON, self._run_operation)
        self._sizer.Add(self._run_button, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self.Layout()
        self.Thaw()

    @property
    def wx_add_options(self) -> Tuple[int, ...]:
        return (1,)

    def _on_mode_change(self, evt):
        self._mode_description.SetLabel(
            MODES[self._mode.GetString(self._mode.GetSelection())]
        )
        self._mode_description.Fit()
        self.Layout()
        evt.Skip()

    def _on_pick_block_button(self, evt):
        """Set up listening for the block click"""
        self._show_pointer = True

    def _on_box_click(self):
        if self._show_pointer:
            self._show_pointer = False
            x, y, z = self._pointer.pointer_base
            self._block_define.universal_block = (
                self.world.get_block(x, y, z, self.canvas.dimension),
                None,
            )

    def _get_fill_block(self):
        return self._block_define.universal_block[0]

    def disable(self):
        self._save_options(
            {
                "fill_block": self._get_fill_block(),
                "fill_block_options": (
                    self._block_define.platform,
                    self._block_define.version_number,
                    self._block_define.force_blockstate,
                    self._block_define.namespace,
                    self._block_define.block_name,
                    self._block_define.properties,
                ),
            }
        )

    def _run_operation(self, _):
        self.canvas.run_operation(lambda: self._waterlog())

    def _waterlog(self):
        mode = self._mode.GetString(self._mode.GetSelection())
        waterlog_block = self._get_fill_block().base_block
        world = self.world
        selection = self.canvas.selection.selection_group
        dimension = self.canvas.dimension
        iter_count = len(list(world.get_coord_box(dimension, selection, True)))
        count = 0
        for chunk, slices, _ in world.get_chunk_slice_box(dimension, selection, True):
            original_blocks = chunk.blocks[slices]
            palette, blocks = numpy.unique(original_blocks, return_inverse=True)
            blocks = blocks.reshape(original_blocks.shape)
            if mode == "Overlay":
                lut = numpy.array(
                    [
                        world.block_palette.get_add_block(
                            waterlog_block
                            if world.block_palette[block_id].namespaced_name
                            == "universal_minecraft:air"
                            else world.block_palette[block_id].base_block
                            + waterlog_block  # get the Block object for that id and add the user specified block
                        )  # register the new block / get the numerical id if it was already registered
                        for block_id in palette
                    ]  # add the new id to the palette
                )
            elif mode == "Underlay":
                lut = numpy.array(
                    [
                        world.block_palette.get_add_block(
                            waterlog_block
                            if world.block_palette[block_id].namespaced_name
                            == "universal_minecraft:air"
                            else waterlog_block
                            + world.block_palette[  # get the Block object for that id and add the user specified block
                                block_id
                            ].base_block
                        )  # register the new block / get the numerical id if it was already registered
                        for block_id in palette
                    ]  # add the new id to the palette
                )
            elif mode == "Normal Fill":
                lut = numpy.array(
                    [
                        world.block_palette.get_add_block(
                            waterlog_block
                        )  # register the new block / get the numerical id if it was already registered
                    ]
                    * len(palette)  # add the new id to the palette
                )
            elif mode == "Game Fill":
                lut = numpy.array(
                    [
                        world.block_palette.get_add_block(
                            waterlog_block + world.block_palette[block_id].base_block
                            if world.block_palette[block_id].namespaced_name
                            == "universal_minecraft:water"
                            else waterlog_block
                        )  # register the new block / get the numerical id if it was already registered
                        for block_id in palette
                    ]  # add the new id to the palette
                )
            elif mode == "Set First":
                lut = numpy.array(
                    [
                        world.block_palette.get_add_block(
                            waterlog_block
                            + world.block_palette[block_id].extra_blocks[0]
                            if world.block_palette[block_id].extra_blocks
                            else waterlog_block
                        )  # register the new block / get the numerical id if it was already registered
                        for block_id in palette
                    ]  # add the new id to the palette
                )
            elif mode == "Set Second":
                lut = numpy.array(
                    [
                        world.block_palette.get_add_block(
                            world.block_palette[block_id].base_block + waterlog_block
                        )  # register the new block / get the numerical id if it was already registered
                        for block_id in palette
                    ]  # add the new id to the palette
                )
            else:
                raise Exception("hello")

            chunk.blocks[slices] = lut[blocks]

            chunk.changed = True
            count += 1
            yield count / iter_count


export = {
    "name": "Waterlog",  # the name of the plugin
    "operation": Waterlog,  # the actual function to call when running the plugin
}



import wx
from concurrent.futures import ThreadPoolExecutor
import webbrowser
from typing import TYPE_CHECKING, Callable

from amulet import load_format
from amulet.api.level import BaseLevel

from amulet_map_editor import lang, log
from amulet_map_editor.api.wx.ui.simple import SimplePanel
from amulet_map_editor.api.wx.ui.select_world import WorldSelectDialog, WorldUI
from amulet_map_editor.api.datatypes import MenuData
from amulet_map_editor.api.framework.programs import BaseProgram

if TYPE_CHECKING:
    from amulet.api.wrapper import WorldFormatWrapper

thread_pool_executor = ThreadPoolExecutor(max_workers=1)
work_count = 0


class ConvertExtension(SimplePanel, BaseProgram):
    def __init__(
        self, container, world: BaseLevel, close_self_callback: Callable[[], None]
    ):
        SimplePanel.__init__(self, container)
        self.world = world
        self._close_self_callback = close_self_callback

        self._close_world_button = wx.Button(
            self, wx.ID_ANY, label=lang.get("world.close_world")
        )
        self._close_world_button.Bind(wx.EVT_BUTTON, self._close_world)
        self.add_object(self._close_world_button, 0, wx.ALL | wx.CENTER)

        self._input = SimplePanel(self, wx.HORIZONTAL)
        self.add_object(self._input, 0, wx.ALL | wx.CENTER)
        self._input.add_object(
            wx.StaticText(
                self._input,
                wx.ID_ANY,
                lang.get("program_convert.input_world"),
                wx.DefaultPosition,
                wx.DefaultSize,
                0,
            ),
            0,
            wx.ALL | wx.CENTER,
        )
        self._input.add_object(
            WorldUI(self._input, self.world.level_wrapper), 0, wx.ALL | wx.CENTER
        )

        self._output = SimplePanel(self, wx.HORIZONTAL)
        self.add_object(self._output, 0, wx.ALL | wx.CENTER)
        self._output.add_object(
            wx.StaticText(
                self._output,
                wx.ID_ANY,
                lang.get("program_convert.output_world"),
                wx.DefaultPosition,
                wx.DefaultSize,
                0,
            ),
            0,
            wx.ALL | wx.CENTER,
        )

        self._select_output_button = wx.Button(
            self, wx.ID_ANY, label=lang.get("program_convert.select_output_world")
        )
        self._select_output_button.Bind(wx.EVT_BUTTON, self._show_world_select)
        self.add_object(self._select_output_button, 0, wx.ALL | wx.CENTER)

        self._convert_bar = SimplePanel(self, wx.HORIZONTAL)
        self.add_object(self._convert_bar, 0, wx.ALL | wx.CENTER)

        self.loading_bar = wx.Gauge(
            self._convert_bar,
            wx.ID_ANY,
            100,
            wx.DefaultPosition,
            wx.DefaultSize,
            wx.GA_HORIZONTAL,
        )
        self._convert_bar.add_object(self.loading_bar, options=wx.ALL | wx.EXPAND)
        self.loading_bar.SetValue(0)

        self.convert_button = wx.Button(
            self._convert_bar,
            wx.ID_ANY,
            label=lang.get("program_convert.convert_button"),
        )
        self._convert_bar.add_object(self.convert_button)
        self.convert_button.Bind(wx.EVT_BUTTON, self._convert_event)

        self.out_world_path = None

    def menu(self, menu: MenuData) -> MenuData:
        menu.setdefault(lang.get("menu_bar.help.menu_name"), {}).setdefault(
            "control", {}
        ).setdefault(
            lang.get("program_convert.menu_bar.help.user_guide"),
            lambda evt: self._help_controls(),
        )
        return menu

    def _help_controls(self):
        webbrowser.open(
            "https://github.com/Amulet-Team/Amulet-Map-Editor/blob/master/amulet_map_editor/programs/convert/readme.md"
        )

    def _show_world_select(self, evt):
        select_world = WorldSelectDialog(self, self._output_world_callback)
        select_world.ShowModal()
        select_world.Destroy()

    def _output_world_callback(self, path):
        if path == self.world.level_path:
            wx.MessageBox(lang.get("program_convert.input_output_must_different"))
            return
        try:
            out_world_format = load_format(path)
            self.out_world_path = path

        except Exception:
            return

        for child in list(self._output.GetChildren())[1:]:
            child.Destroy()
        self._output.add_object(WorldUI(self._output, out_world_format), 0)
        self._output.Layout()
        self._output.Fit()
        self.Layout()
        # self.Fit()

    def _update_loading_bar(self, chunk_index, chunk_total):
        wx.CallAfter(self.loading_bar.SetValue, int(100 * chunk_index / chunk_total))

    def _convert_event(self, evt):
        if self.out_world_path is None:
            wx.MessageBox(lang.get("program_convert.select_before_converting"))
            return
        self.convert_button.Disable()
        global work_count
        work_count += 1
        thread_pool_executor.submit(self._convert_method)
        # self.world.save(self.out_world, self._update_loading_bar)

    def _convert_method(self):
        global work_count
        try:
            out_world = load_format(self.out_world_path)
            log.info(f"Converting world {self.world.level_path} to {out_world.path}")
            out_world: WorldFormatWrapper
            out_world.open()
            self.world.save(out_world, self._update_loading_bar)
            out_world.close()
            message = lang.get("program_convert.conversion_completed")
            log.info(
                f"Finished converting world {self.world.level_path} to {out_world.path}"
            )
        except Exception as e:
            message = f"Error during conversion\n{e}"
            log.error(message, exc_info=True)
        self._update_loading_bar(0, 100)
        self.convert_button.Enable()
        wx.MessageBox(message)
        work_count -= 1

    def is_closeable(self):
        if work_count:
            log.info(
                f"World {self.world.level_path} is still being converted. Please let it finish before closing"
            )
        return work_count == 0

    def _close_world(self, evt):
        self._close_self_callback()



from typing import TYPE_CHECKING, Tuple
import wx

from amulet.operations.fill import fill

from amulet_map_editor.api.wx.ui.base_select import EVT_PICK
from amulet_map_editor.api.wx.ui.block_select import BlockDefine
from amulet_map_editor.programs.edit.api.operations import DefaultOperationUI

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class Fill(wx.Panel, DefaultOperationUI):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        wx.Panel.__init__(self, parent)
        DefaultOperationUI.__init__(self, parent, canvas, world, options_path)
        self.Freeze()
        self._sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self._sizer)

        options = self._load_options({})

        self._block_define = BlockDefine(
            self,
            world.translation_manager,
            wx.VERTICAL,
            *(options.get("fill_block_options", []) or [world.level_wrapper.platform]),
            show_pick_block=True
        )
        self._block_define.Bind(EVT_PICK, self._on_pick_block_button)
        self._sizer.Add(self._block_define, 1, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self._run_button = wx.Button(self, label="Run Operation")
        self._run_button.Bind(wx.EVT_BUTTON, self._run_operation)
        self._sizer.Add(self._run_button, 0, wx.ALL | wx.ALIGN_CENTRE_HORIZONTAL, 5)

        self.Layout()
        self.Thaw()

    @property
    def wx_add_options(self) -> Tuple[int, ...]:
        return (1,)

    def _on_pick_block_button(self, evt):
        """Set up listening for the block click"""
        self._show_pointer = True

    def _on_box_click(self):
        if self._show_pointer:
            self._show_pointer = False
            x, y, z = self._pointer.pointer_base
            self._block_define.universal_block = (
                self.world.get_block(x, y, z, self.canvas.dimension),
                None,
            )

    def _get_fill_block(self):
        return self._block_define.universal_block[0]

    def disable(self):
        self._save_options(
            {
                "fill_block": self._get_fill_block(),
                "fill_block_options": (
                    self._block_define.platform,
                    self._block_define.version_number,
                    self._block_define.force_blockstate,
                    self._block_define.namespace,
                    self._block_define.block_name,
                    self._block_define.properties,
                ),
            }
        )

    def _run_operation(self, _):
        self.canvas.run_operation(
            lambda: fill(
                self.world,
                self.canvas.dimension,
                self.canvas.selection.selection_group,
                self._get_fill_block(),
            )
        )


export = {
    "name": "Fill",  # the name of the plugin
    "operation": Fill,  # the actual function to call when running the plugin
}



import wx
import os
from typing import Type, Union

from amulet_map_editor.api.resources import (
    ResourceItem,
    ResourceDirectory,
    MissingResourceItem,
)


class BitmapResourceItem(ResourceItem):
    def bitmap(self, width=-1, height=-1, quality=wx.IMAGE_QUALITY_HIGH) -> wx.Bitmap:
        bm = wx.Bitmap(self._path)

        if width > 0 or height > 0:
            im: wx.Image = bm.ConvertToImage()
            im = im.Scale(
                width if width > 0 else bm.Width,
                height if height > 0 else bm.Height,
                quality,
            )
            bm = im.ConvertToBitmap()
        return bm

    @classmethod
    def default_object(cls) -> wx.Bitmap:
        bm = wx.Bitmap()
        bm.Create(32, 32)
        return bm


class ImageResourceDirectory(ResourceDirectory):
    ALLOWED_EXTS = (".png",)

    def __init__(
        self, directory_name: str, factory_class: Type[ResourceItem], parent: str
    ):
        super().__init__(directory_name, factory_class, parent)

    def __getattr__(
        self, item
    ) -> Union["ImageResourceDirectory", BitmapResourceItem, MissingResourceItem]:
        return super().__getattr__(item)


_self = ImageResourceDirectory(
    "image", BitmapResourceItem, os.path.dirname(os.path.dirname(__file__))
)


def __getattr__(
    name: str,
) -> Union["ImageResourceDirectory", BitmapResourceItem, MissingResourceItem]:
    return _self.__getattr__(name)


def __dir__():
    return _self.__dir__()


TABLERICONS: ImageResourceDirectory = _self.icon.tablericons

ADD_ICON: BitmapResourceItem = TABLERICONS.plus_green
SUBTRACT_ICON: BitmapResourceItem = TABLERICONS.minus_red
EDIT_ICON: BitmapResourceItem = TABLERICONS.edit
REFRESH_ICON: BitmapResourceItem = TABLERICONS.refresh

UP_ARROW: BitmapResourceItem = TABLERICONS.chevron_up
DOWN_ARROW: BitmapResourceItem = TABLERICONS.chevron_down

UP_CARET: BitmapResourceItem = TABLERICONS.caret_up
DOWN_CARET: BitmapResourceItem = TABLERICONS.caret_down

COLOUR_PICKER: BitmapResourceItem = TABLERICONS.color_picker
TRASH: BitmapResourceItem = TABLERICONS.trash
MAXIMIZE: BitmapResourceItem = TABLERICONS.arrows_maximize
MINIMIZE: BitmapResourceItem = TABLERICONS.arrows_minimize



#!/usr/bin/env python3

if __name__ == "__main__":
    try:
        import sys

        if sys.version_info[:2] < (3, 7):
            raise Exception("Must be using Python 3.7+")

        import os
        import traceback
        import wx
        from amulet_map_editor import log
        from amulet_map_editor.api.framework import AmuletApp

        if sys.platform == "linux" and wx.VERSION >= (4, 1, 1):
            # bug 247
            os.environ["PYOPENGL_PLATFORM"] = "egl"

    except Exception as e:
        err = f"Failed to import requirements. Check that you extracted correctly.\n{e}"
        print(err)
        with open("./logs/amulet_map_editor_.log") as f:
            f.write(err)
        input("Press ENTER to continue.")
    else:
        try:
            app = AmuletApp(0)
            app.MainLoop()
        except Exception as e:
            log.critical(
                f"Amulet Crashed. Sorry about that. Please report it to a developer if you think this is an issue. \n{traceback.format_exc()}"
            )
            input("Press ENTER to continue.")



import webbrowser
import wx
import wx.lib.inspection

from amulet_map_editor.api import image, lang
from .base_page import BasePageUI
from amulet_map_editor.api.wx.ui.select_world import WorldSelectDialog


class AmuletMainMenu(wx.Panel, BasePageUI):
    def __init__(self, parent: wx.Window, open_world):
        super(AmuletMainMenu, self).__init__(parent)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)
        sizer.AddStretchSpacer(1)
        self._open_world_callback = open_world
        name_sizer = wx.BoxSizer()
        sizer.Add(name_sizer, 0, wx.CENTER)
        icon_img = image.logo.amulet_logo.bitmap(64, 64)

        icon = wx.StaticBitmap(self, wx.ID_ANY, icon_img, (0, 0), (64, 64))
        icon2 = wx.StaticBitmap(self, wx.ID_ANY, icon_img, (0, 0), (64, 64))
        icon2.Bind(
            wx.EVT_LEFT_DOWN, lambda evt: wx.lib.inspection.InspectionTool().Show()
        )
        name_sizer.Add(icon, flag=wx.CENTER)

        amulet_name = wx.StaticText(self, label="Amulet")
        amulet_name.SetFont(wx.Font(40, wx.DECORATIVE, wx.NORMAL, wx.NORMAL))
        name_sizer.Add(amulet_name, flag=wx.CENTER | wx.LEFT | wx.RIGHT, border=10)
        name_sizer.Add(icon2, flag=wx.CENTER)
        button_font = wx.Font(20, wx.DECORATIVE, wx.NORMAL, wx.NORMAL)
        self._open_world_button = wx.Button(
            self, label=lang.get("main_menu.open_world"), size=(400, 70)
        )
        self._open_world_button.SetFont(button_font)
        self._open_world_button.Bind(wx.EVT_BUTTON, self._show_world_select)
        sizer.Add(self._open_world_button, 0, wx.ALL | wx.CENTER, 5)

        self._help_button = wx.Button(
            self, label=lang.get("main_menu.help"), size=(400, 70)
        )
        self._help_button.SetToolTip(lang.get("app.browser_open_tooltip"))
        self._help_button.SetFont(button_font)
        self._help_button.Bind(wx.EVT_BUTTON, self._documentation)
        sizer.Add(self._help_button, 0, wx.ALL | wx.CENTER, 5)

        self._discord_button = wx.Button(
            self, label=lang.get("main_menu.discord"), size=(400, 70)
        )
        self._discord_button.SetToolTip(lang.get("app.browser_open_tooltip"))
        self._discord_button.SetFont(button_font)
        self._discord_button.Bind(wx.EVT_BUTTON, self._discord)
        sizer.Add(self._discord_button, 0, wx.ALL | wx.CENTER, 5)

        sizer.AddStretchSpacer(2)

    def _show_world_select(self, _):
        select_world = WorldSelectDialog(self, self._open_world_callback)
        select_world.ShowModal()
        select_world.Destroy()

    @staticmethod
    def _documentation(_):
        webbrowser.open(
            "https://github.com/Amulet-Team/Amulet-Map-Editor/blob/master/amulet_map_editor/readme.md"
        )

    @staticmethod
    def _discord(_):
        webbrowser.open("https://discord.gg/BTm6jnf")

    def enable(self):
        self.GetGrandParent().create_menu()



from typing import TYPE_CHECKING
import wx

from amulet.api.selection import SelectionGroup
from amulet.api.data_types import Dimension, OperationReturnType

from amulet_map_editor.programs.edit.api.operations import (
    SimpleOperationPanel,
    OperationSilentAbort,
)

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel
    from amulet_map_editor.programs.edit.api.canvas import EditCanvas


class Clone(SimpleOperationPanel):
    def __init__(
        self,
        parent: wx.Window,
        canvas: "EditCanvas",
        world: "BaseLevel",
        options_path: str,
    ):
        SimpleOperationPanel.__init__(self, parent, canvas, world, options_path)
        self._add_run_button()
        self.Layout()

    def _operation(
        self, world: "BaseLevel", dimension: Dimension, selection: SelectionGroup
    ) -> OperationReturnType:
        structure = world.extract_structure(selection, dimension)
        self.canvas.paste(structure, structure.dimensions[0])
        raise OperationSilentAbort


export = {
    "name": "Clone",  # the name of the plugin
    "operation": Clone,  # the actual function to call when running the plugin
}



from typing import List
from setuptools import setup, find_packages
import os
import glob
import shutil
import sys

try:
    import versioneer
except ImportError:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)))
    import versioneer

# there were issues with other builds carrying over their cache
for d in glob.glob("*.egg-info"):
    shutil.rmtree(d)


def load_requirements(path: str) -> List[str]:
    requirements = []
    with open(path) as f:
        for line in f.readlines():
            line = line.strip()
            if line.startswith("git+") or line.startswith("https:"):
                continue
            elif line.startswith("-r "):
                requirements += load_requirements(line[3:])
            else:
                requirements.append(line)
    return requirements


required_packages = load_requirements("./requirements.txt")

try:
    # try and freeze the requirements if already installed
    from pip._internal.operations import freeze

    first_party = {
        "amulet-core",
        "amulet-nbt",
        "pymctranslate",
        "minecraft-resource-pack",
    }
    installed = {r.split("==")[0].lower(): r for r in freeze.freeze() if "==" in r}
    for index, r in enumerate(required_packages):
        if r[0] != "#" and "~=" in r:
            req = r.split("~=")[0]
            if req in first_party and req in installed:
                required_packages[index] = installed[req]
except:
    pass

package_data = [
    os.path.relpath(path, "amulet_map_editor")
    for path in set(
        glob.glob(os.path.join("amulet_map_editor", "**", "*.*"), recursive=True)
    )
    - set(
        glob.glob(os.path.join("amulet_map_editor", "**", "*.py[cod]"), recursive=True)
    )
]

setup(
    install_requires=required_packages,
    packages=find_packages(),
    package_data={"amulet_map_editor": package_data},
    cmdclass=versioneer.get_cmdclass(),
)



from .camera import (
    Camera,
    CameraMovedEvent,
    EVT_CAMERA_MOVED,
    ProjectionChangedEvent,
    EVT_PROJECTION_CHANGED,
    Projection,
)
from .controllable_camera import ControllableCamera
